#
# Copyright (c) 2023, HILSTER - https://hilster.io
# All rights reserved.
#

from typing import Optional, Union, Type, Callable, List

from .interface import HartInterface
from .application_layer import HartApplicationLayer
import datetime
import logging
import htf.hart
from htf.hart.protocol import HartFrame
import argparse
import sys

logger = logging.getLogger(__name__)


class HartMonitor:
    """
    A HART Monitor that can be used interactively.
    """

    def __init__(
        self,
        comport: Optional[str] = None,
        decoder_generator: Optional[Union[Type[HartFrame], Callable[..., HartFrame]]] = None,
        protocol_factory: Optional[Callable[..., Type[HartFrame]]] = None,
        interactive: bool = False,
        argv: Optional[List[str]] = None,
    ):
        """
        Args:
            comport: the comport to be used where the HART modem is
                connected to.
            decoder_generator=None: a callable that returns
                a HartFrame instance with or without device specific commands that
                is used to decode all frames (if ``protocol_factory`` is ``None``).
            protocol_factory=None: a protocol factory that returns
                a device specific HART frame parser with respect to the
                supplied device address. For every received frame the frame is
                decoded using the ``decoder_generator`` that may be ``None``
                and after that a device specific decoder is created respecting
                the device address.
            interactive=True: if set to ``True`` the HART Monitor becomes
                interactive
            argv=None: used for testing, etc.
        """
        if argv is None:
            argv = sys.argv

        self._interface = None
        self._application_layer = None
        self._comport = comport
        self._show_version = False
        self._database = None
        self._protocol_factory = protocol_factory
        self._interactive = interactive

        self.set_protocol_factory(protocol_factory)

        if self._interactive is True:
            self._parse_arguments(argv)

        self._interface = HartInterface(self._comport)  # type: ignore
        self._application_layer = HartApplicationLayer(interface=self._interface, decoder_generator=decoder_generator)

        self.set_decoder_generator(decoder_generator)

        if self._interactive is True:
            self._run_interactively()

    def close(self) -> None:
        if self._application_layer is not None:
            self._application_layer.close()

    def __del__(self) -> None:
        self.close()

    def _parse_arguments(self, argv: List[str]) -> None:
        self._main_parser = argparse.ArgumentParser(
            description="hart-monitor - HILSTER Testing Framework (HILSTER GmbH, https://hilster.io)"
        )

        self._main_parser.add_argument(
            "-c",
            "--comport",
            default=self._comport,
            dest="_comport",
            metavar="COMPORT",
            required=True,
            help="comport for HART communication",
        )

        self._main_parser.add_argument(
            "-d",
            "--database",
            default=None,
            dest="_database",
            metavar="Protocol database",
            help="An import string to the protocol database where device specific "
            "HART frames can be found. See docs for details.",
        )

        self._main_parser.add_argument(
            "-V",
            "--version",
            default=self._comport,
            dest="_show_version",
            action="store_true",
            help="Show version and exit",
        )

        self._main_parser.parse_args(argv[1:], self)

        if self._show_version:
            print("hart-monitor-" + htf.hart.__version__)
            exit(0)

        if self._database is not None:
            print("Loading protocol database from", self._database)
            try:
                parts = self._database.split(".")
                import_string = ".".join(parts[:-1])
                logger.info("Importing database %s" % self._database)
                module = __import__(import_string)
                for part in parts[1:-1]:
                    module = getattr(module, part)
                protocol_factory = getattr(module, parts[-1])
                self.set_protocol_factory(protocol_factory)
            except ImportError:
                logger.info("Could not load database from '%s'" % self._database)
        # exit(0)

    def _run_interactively(self) -> None:
        self.run(0)
        self.close()
        exit(0)

    def set_protocol_factory(self, protocol_factory: Optional[Callable[..., Type[HartFrame]]]) -> None:
        """
        Set the protocol factory.

        Args:
            protocol_factory=None: a protocol factory that returns
                a device specific HART frame parser with respect to the
                supplied device address.
        """
        self._protocol_factory = protocol_factory

    def set_decoder_generator(
        self, decoder_generator: Optional[Union[Type[HartFrame], Callable[..., HartFrame]]] = None
    ) -> None:
        """
        Set the decoder generator. If ``decoder_generator`` is ``None``
        ``HartFrame`` is used.

        Args:
            decoder_generator=HartFrame: a parser for HART frames.
        """
        if decoder_generator is None:
            decoder_generator = HartFrame
        self._decoder_generator = decoder_generator
        self._application_layer.set_decoder_generator(decoder_generator)  # type: ignore

    def _parse_frame(self, frame: HartFrame) -> HartFrame:
        if frame.command.get() == 0:
            return frame
        if self._protocol_factory is None:
            return frame
        # a protocol factory was supplied
        decoder = self._protocol_factory(frame.address.address.get())()
        decoder.decode(frame.encode())
        return decoder

    def handle_frame(self, frame: HartFrame) -> None:  # pragma: no cover
        """
        The frame handler that prints the formatted output.
        This method can be overwritten.

        Args:
            frame: the HART frame to be monitored.
        """
        print("-" * 80)
        print(datetime.datetime.utcnow())
        # print(frame.introspect())
        print(frame)
        logger.info(str(frame))

    def handle_exception(self, frame: Optional[HartFrame], exc: BaseException) -> None:
        """
        The exception handler that prints the formatted output.
        This method can be overwritten.

        This overwritten method may or may not call ``raise`` at
        the end to raise the exception again.

        Args:
            frame: the HART frame to be monitored.
            exc: the exception that can be formatted.
        """
        print("An exception was raised:", exc)
        # raise

    def run(self, frame_count: int = 0) -> None:
        """
        Run the monitor and display received HART frames.

        Args:
            frame_count: The number of frames to be received. ``0`` means forever.
        """
        if frame_count == 0:

            def condition(x: int) -> bool:
                return True

        else:

            def condition(x: int) -> bool:
                return x <= frame_count

        i = 0
        while condition(i):
            frame: Optional[HartFrame] = None
            try:
                frame = self._application_layer.read()  # type: ignore
                if frame is not None:
                    frame = self._parse_frame(frame)
                    self.handle_frame(frame)
            except KeyboardInterrupt:
                print("Keyboard interrupt was received. Stopping.")
                return
            except BaseException as e:
                self.handle_exception(frame, e)
            i += 1
